#! /usr/bin/tclsh
#
#
#######################################################################################################
#
#	Tcl script v1.0 implementing APRSMSG, a Web to APRS Message system.
#	Written by Max Monti ik3svw@ir3ip.net http://www.ir3ip.net/ik3svw - sep 2004
#
#	Warning about aprsmsg system: if you let others to send messages wia web aprs message page,
#	then they can send packets to internet using your main Call (SysCall) as set in this tcl script.
#	So you are still responsible for any packet sent with your main Call thru this tcl script even
#	if someone else is sending packets unsing your web aprs message page (aprsmsg html/cgi).
#
#######################################################################################################
#
#
#
#			Please change main parameters below to let this script to work correctly.
#
###################### START - MAIN PARAMETERS TO BE CHANGED WITH APPROPIATE VALUES ##################
#
set SysCall "NOCALL";				# Main Callsign.
set SysPass "00000"; 				# Aprsd validation number for Main callsign.
#
set ServerIp "000.000.000.000";		# Ip address of aprsd server.
set ServerPort "10153";				# Port number of aprsd server.
#
set APRSMSG_FILE "./Aprs_msg";		# Path and name of aprs_msg file from cgi interface.
set MHEARD_FILE "./www/MHeard.js";		# Path and name of .js file to be read by web page.
set EXTLOG_FILE "./Output.log";		# Path and name of Log file.
#
set MyBeacon "=0000.00N\\00000.00E-AprsMsg System"; # APRS Location beacon string (note \\ instead of \).
set MyStatus ">http://www.WebServer.net/aprsmsghelp.html for info"; # APRS Status beacon string.
#
set mheard_interval 600;	# Time in seconds for updating MHeard file used in web page. Suggested value 600.
set beacon_interval 1200;	# Time in seconds for sending APRS location beacon. Suggested value 1200.
set status_interval 1800;	# Time in seconds for sending APRS status beacon. Suggested value 1800.
#
####################### END - MAIN PARAMETERS TO BE CHANGED WITH APPROPIATE VALUES ###################
#
#
#
#
#################### The following code has no user settings - do not change it ###################
#
#
############## Checks for external messages and queue them in FIFO list ###########
############## This procedure has no user settings - do not change code ###########
proc CheckExtMsg {} {
	global PktQueue MHeard SysCall APRSMSG_FILE fdt
	# Try to open message file made by cgi interface
	if {![catch {open $APRSMSG_FILE RDONLY} fdm]} {
		while {! [eof $fdm]} {
			gets $fdm msgline
			if {"$msgline" != ""} {
				# Parse message line and retreive message data
				set msgline [split "$msgline" "|"]
				# Check if receiver call is in MHeard list
				set Hidx [IsHeard [lindex $msgline 0]]
				if {$Hidx != -1} {
				# Format aprs message
				set Tris [split [lindex $MHeard $Hidx] "|"]
				set Ts [string repeat " " [expr 9 - [string length [lindex $Tris 0]]]]
				set NewPkt "$SysCall>APRS::[lindex $Tris 0]$Ts:[lindex $msgline 1]"
				# Check if message already queued
				if {[lsearch -exact $PktQueue $NewPkt] == -1} {
					# Retreive sender call
					set Bis [split [lindex $msgline 1] ">"]
					# Write action to Log file
					puts $fdt "Sending APRS Msg to [lindex $Tris 0] from [lindex $Bis 0] on [clock format [clock seconds]]"
					flush $fdt
					# Queue message
					lappend PktQueue $NewPkt
					}
				}
			}
		}
		close $fdm
		# Message file was processed, so delete it
		file delete $APRSMSG_FILE
	}
}
#
#
################## Sends a packet from FIFO APRS queue ###############
###### This procedure has no user settings - do not change code ######
proc CheckPktQueue {} {
	global sd PktQueue
	if {[llength $PktQueue] == 0} {return}
	# Send to socket aprs packet rertieved from queue
	puts $sd "[lindex "$PktQueue" 0]\r"
	flush $sd
	set PktQueue [lreplace "$PktQueue" 0 0]
}
#
#
######################### Update MHeard List #########################
###### Update MHeard only if station may have message capability #####
###### This procedure has no user settings - do not change code ######
proc UpdateMHeard {Call Path Table} {
	global MHeard
	# Check for message capability
	if { $Table == "=" || $Table == "@" || $Table == "'" || $Table == "`" || $Table == ":" || $Table == "?" } {
		set N "$Call|[clock seconds]|$Path"
		# Check if call already in MHeard list
		set L [lsearch -glob $MHeard "*$Call|*"]
		if {$L < 0} {
			# Insert new call in MHeard List
			lappend MHeard $N
		} else {
			# Call already exists in MHeard list, so replace entry
			set MHeard [lreplace $MHeard $L $L $N]
	 	}
	}
}
#
#
######################## Delete old MHeard entires ###################
################## Also write on disk MHeard list file ###############
###### This procedure has no user settings - do not change code ######
proc CheckMHeard {tsk} {
	global MHeard MHEARD_FILE
	set MHeardLimit 3600; # Elements older than this seconds will be deleted
	set i [expr [llength $MHeard] - 1]
	# Check elements loop
	while {$i >= 0} {
		set M [lindex $MHeard $i]
		if {[expr $tsk - [lindex [split "$M" "|"] 1]] > $MHeardLimit} {
			set MHeard [lreplace $MHeard $i $i]
		}
	incr i -1
	}
	# Create JavaScript file for use with html message page
	set fd [open $MHEARD_FILE {WRONLY CREAT TRUNC} 0644]
	puts $fd "<!--\nvar MHeard = \"$MHeard\";\nvar MHtime = \"$tsk\";\n//-->"
	close $fd
}
#
#
############## Function to see if a Call is in MHeard List ###########
###### This procedure has no user settings - do not change code ######
proc IsHeard {Call} {
	global MHeard
	return [lsearch -glob $MHeard "*$Call|*"]
}
#
#
#
#
########################################## Main Program ###########################################
#################### The following code has no user settings - do not change it ###################
fconfigure stdin -blocking 0; # Nonblocking stdin
set MHeard "";			# Monitored Heard Callsign List.
set PktQueue "";			# Buffer of APRS packets to send.
set tsc [clock seconds];	# Timer counter for cheking APRS out-List and MHeard List.
set tsb $tsc;			# Timer counter for APRS Beacon Interval.
set tss $tsc;			# Timer counter for APRS Status Interval.
set tsm $tsc;			# Timer counter for MHeard Update.
set scacnt 0;			# Socket connection attempts counter.
set APRS_Beacon "$SysCall>APRS:$MyBeacon";	# Full APRS Location beacon string.
set APRS_Status "$SysCall>APRS:$MyStatus";	# Full APRS Status beacon string.
# Open then log file
set fdt [open $EXTLOG_FILE {WRONLY CREAT TRUNC}]
# Loop for socket attempts - up to 10 times
while { $scacnt < 10 } {
	# Try to open Socket
	if { [catch {set sd [socket "$ServerIp" "$ServerPort"]} ft] } {
		if { $scacnt == 0 } {
		# Not able to open socket, write the result to Log file
		puts $fdt "\nError at [clock format [clock seconds]] due to $ft"
		flush $fdt
		}
		incr scacnt
	} else {
		# Socket was successfully opened			
		puts $sd "user $SysCall pass $SysPass version TCL script by ik3svw\r"; #APRSD server registration
		flush $sd
		puts $fdt "\nProgram Started on [clock format [clock seconds]]"
		flush $fdt
		set scacnt 0
		# Main Program Loop
		while {[gets $sd line]>=0} {
			# Check for updating MHeard Calls list
			if {[regexp {^([0-9A-Z-]+)>[^,]+,?([^:]*):(.).*} $line Pk C P T]} {
				UpdateMHeard $C $P $T
			}
			set tnow [clock seconds]
			# Every 30 seconds check if there is any message from cgi interface
			# and sends one packet rerieved from queue to socket stream
			if {[expr $tnow - $tsc] > 30} {
				set tsc $tnow
				CheckExtMsg
				CheckPktQueue
			}
			# Every mheard interval delete old entires in MHeard list
			# and write it to disk for usage in message html page
			if {[expr $tnow - $tsm] > $mheard_interval} {
				set tsm $tnow
				CheckMHeard $tnow
			}
			# Every beacon interval send APRS location beacon
			if {[expr $tnow - $tsb] > $beacon_interval} {
				set tsb $tnow
				puts $sd "$APRS_Beacon\r"
				flush $sd
			}
			# Every status interval send APRS status beacon
			if {[expr $tnow - $tss] > $status_interval} {
				set tss $tnow
				puts $sd "$APRS_Status\r"
				flush $sd
			}
		}
		if {[eof $sd]} {
			# If socket was closed write the result on Log file
			close $sd
			puts $fdt "Socket closed at [clock format [clock seconds]]"
			puts -nonewline $fdt "Waiting for socket connection"
			flush $fdt
		}
	}
	puts -nonewline $fdt "."
	flush $fdt
	# Wait one minute before attempting to connect socket again
	after 60000
}
# Exit program if unsuccessful connection attempts
puts $fdt "\nToo much socket connection attempts."
puts $fdt "Closing down $SysCall at [clock format [clock seconds]]"
close $fdt
exit
#